//-< Mutex.java >----------------------------------------------------*--------*
// GOODS                      Version 2.02       (c) 1998  GARRET    *     ?  *
// (Generic Object Oriented Database System)                         *   /\|  *
// Java Program Interface                                            *  /  \  *
//                          Created:     01-Feb-2002  K.A. Knizhnik  * / [] \ *
//                          Last update: 01-Feb-2002  K.A. Knizhnik  * GARRET *
//-------------------------------------------------------------------*--------*
// Mutual threads exclusion
//-------------------------------------------------------------------*--------*
namespace Goods
{
    using System;
	
    /// <summary>Class for providing mutual exclusion of threads. Only one thread can enter
    /// critical section guarded by mutex (but can do it several times).
    /// </summary>
    public class Mutex
    {
        /// <summary>Lock mutex. This method should be called before entering critical 
        /// section.
        /// </summary>
        public virtual void  enter()
        {
            lock(this)
            {
                System.Threading.Thread self = System.Threading.Thread.CurrentThread;
                if (owner != null || nBlocked != 0)
                {
                    if (owner == self)
                    {
                        nested += 1;
                        return ;
                    }
                    do 
                    {
                        try
                        {
                            nBlocked += 1;
                            System.Threading.Monitor.Wait(this);
                            nBlocked -= 1;
                        }
                        catch (System.Threading.ThreadInterruptedException ex)
                        {
                            // It is possible for a thread to be interrupted after 
                            // being notified but before returning from the wait() 
                            // call. To prevent lost of notification notify()
                            // is invoked. 
                            System.Threading.Monitor.Pulse(this);
                            nBlocked -= 1;
                            throw ex;
                        }
                    }
                    while (owner != null);
                }
                System.Diagnostics.Debug.Assert(nested == 0);
                owner = self;
                nested = 1;
            }
        }
		
        /// <summary>Try to lock mutex within specified period of time. This method 
        /// should be called before entering critical section.
        /// 
        /// </summary>
        /// <param name="timeout">the maximum time to wait in milliseconds.
        /// </param>
        /// <returns><code>true</code> if mutex is successfully locked, 
        /// <code>false</code> if <code>enter()</code> was terminated 
        /// due to timeout expiration.
        /// 
        /// </returns>
        public virtual bool enter(long timeout)
        {
            lock(this)
            {
                System.Threading.Thread self = System.Threading.Thread.CurrentThread;
                if (owner != null || nBlocked != 0)
                {
                    if (owner == self)
                    {
                        nested += 1;
                        return true;
                    }
                    if (timeout == 0)
                    {
                        return false;
                    }
                    long startTime = DateTime.Now.Ticks;
                    do 
                    {
                        long currentTime = DateTime.Now.Ticks;
                        if (currentTime - startTime >= timeout)
                        {
                            return false;
                        }
                        try
                        {
                            nBlocked += 1;
                            System.Threading.Monitor.Wait(this, TimeSpan.FromTicks(timeout*1000 - currentTime + startTime));
                            nBlocked -= 1;
                        }
                        catch (System.Threading.ThreadInterruptedException ex)
                        {
                            // It is possible for a thread to be interrupted after 
                            // being notified but before returning from the wait() 
                            // call. To prevent lost of notification notify()
                            // is invoked. 
                            System.Threading.Monitor.Pulse(this);
                            nBlocked -= 1;
                            throw ex;
                        }
                    }
                    while (owner != null);
                }
                System.Diagnostics.Debug.Assert(nested == 0);
                owner = self;
                nested = 1;
                return true;
            }
        }
		
        /// <summary>Release mutex. This method should be called after exit from critical 
        /// section. Mutex will be unlocked only if number of <code>leave()</code>
        /// invocations is equal to the number of <code>enter()</code> invocations.
        /// </summary>
        public virtual void  leave()
        {
            lock(this)
            {
                if (System.Threading.Thread.CurrentThread != owner)
                {
                    throw new NotOwnerError();
                }
                System.Diagnostics.Debug.Assert(nested > 0);
                if (--nested == 0)
                {
                    owner = null;
                    if (nBlocked != 0)
                    {
                        System.Threading.Monitor.Pulse(this);
                    }
                }
            }
        }
		
        protected internal System.Threading.Thread owner;
        protected internal int nested;
        protected internal int nBlocked;
    }
}