//-< Persistent.java >-----------------------------------------------*--------*
// GOODS                      Version 2.02       (c) 1998  GARRET    *     ?  *
// (Generic Object Oriented Database System)                         *   /\|  *
// Java Program Interface                                            *  /  \  *
//                          Created:      1-Oct-98    K.A. Knizhnik  * / [] \ *
//                          Last update: 16-Nov-98    K.A. Knizhnik  * GARRET *
//-------------------------------------------------------------------*--------*
// Base class for all persistent objects
//-------------------------------------------------------------------*--------*
namespace Goods
{
    using System;

    [Flags]
    internal enum ObjectState 
    {
        DIRTY = 0x0001, // object was changed
        RAW = 0x0002, // object has to be loaded from server
        CACHED = 0x0004, // object is placed in LRU cache
        DESTRUCTED = 0x0008, // object was thrown away
        TRANSREAD = 0x0010, // object was accessed for reading in transaction
        TRANSWRITE = 0x0020, // object was accessed for writing in transaction
        USEFUL = 0x0040, // object was access at least twice not sequentially
        ACCESSED = 0x0080, // object was accessed at least once
        SLOCKED = 0x0100, // object is shared locked
        XLOCKED = 0x0200, // object is exslusivly locked
        NEW = 0x0400, // object was just created
        OPTIMISTIC = 0x0800, // object under optimistic metaobject control and has to be validated at sserver
        NOTIFY = 0x1000  // notification about object change should be send
    };

    public class ObjectHeader 
    {
        internal Metaobject metaobject;
		
        internal ClassDescriptor desc;
		
        internal bool invalidated;
        internal byte[] data; // packed object data (as in the storage) 
		
        internal int opid;
        internal Storage storage;
		
        internal ObjectHeader next;
        internal ObjectHeader prev;
		
        internal int accessCount;
        internal ObjectState state;
        internal Persistent obj;

        internal ObjectHeader(Persistent obj) 
        { 
            this.obj = obj;
        }

        internal int size() 
        { 
            return obj._size(desc);
        }

        internal void reset() 
        { 
            obj._reset(desc);
        }

        internal void unpack(ObjectHeader[] refs, int nRefs, int varyingLength) 
        {
            obj._unpack(desc, refs, nRefs, varyingLength);
        }

        internal void pack(byte[] buf, int offs) 
        {
            obj._pack(desc, buf, offs);
        }
    }
	
    /// <summary> Root class for all persistent capable objects
    /// </summary>
    public class Persistent : ContextBoundObject
    {
        public virtual Database Database
        {
            get
            {
                return hdr.storage != null ? hdr.storage.database : null;
            }
			
        }
        public virtual int Oid
        {
            get
            {
                return hdr.opid;
            }
			
        }
        public virtual int StorageId
        {
            get
            {
                return hdr.storage != null ? hdr.storage.id : -1;
            }
			
        }
		
        /// <summary>
        /// Mark object as ben modified. Prorgammer should use this method if he(she) updates the object.
        /// </summary>

        public void Modify()
        {
            hdr.state |= ObjectState.DIRTY;
        }

		
        public Metaobject Mop 
        { 
            get 
            { 
                return hdr.metaobject;
            }
        }

        ~Persistent()
        {
            if (hdr.opid != 0)
            {
                Metaobject mo = hdr.metaobject;
                if (mo != null) 
                { 
                    mo.forgetObject(hdr);
                } 
            }
        }
		
        public virtual void  attachToStorage(Database db, int sid)
        {
            lock(this)
            {
                System.Diagnostics.Debug.Assert(hdr.opid == 0, "persistent object can't change its location");
                hdr.storage = db.storages[sid];
            }
        }
		
        public virtual void  clusterWith(Persistent obj)
        {
            lock(this)
            {
                System.Diagnostics.Debug.Assert(hdr.opid == 0, "persistent object can't change its location");
                hdr.storage = obj.hdr.storage;
            }
        }
		
		
        /// <summary> Get the dabase assigned id of the object
        /// </summary>		
        public override int GetHashCode()
        {
            System.Diagnostics.Debug.Assert(hdr.opid != 0);
            return hdr.opid;
        }
		
        protected internal Persistent()
        {
            hdr = new ObjectHeader(this);
            hdr.desc = ClassDescriptor.lookup(GetType());
            hdr.metaobject = hdr.desc.defaultMetaobject;
            hdr.state = ObjectState.NEW;
            //metaobject.preDaemon(this, Metaobject.MUTATOR|Metaobject.CONSTRUCTOR);
        }

        protected internal Persistent(Metaobject mop)
        {
            hdr = new ObjectHeader(this);
            hdr.metaobject = mop;
        }

        internal ObjectHeader hdr;

        internal ObjectHeader _hdr() 
        {
            return hdr;
        }

        internal void _pack(ClassDescriptor desc, byte[] buf, int offs) 
        {
            desc.pack(this, buf, offs);
        }

        internal void _unpack(ClassDescriptor desc, ObjectHeader[] refs, int nRefs, int varyingLength) 
        {
            desc.unpack(this, refs, nRefs, varyingLength);
        }

        internal int _size(ClassDescriptor desc) 
        {
            return desc.objectSize(this);
        }  
 
        internal void _reset(ClassDescriptor desc) 
        {
            desc.reset(this);
        }
    }
}