#include <assert.h>
#include <stdlib.h>
#include <string.h>
#include "goods.h"
#include "dbscls.h"

static const nat4 kMaxBugsPerLeaf     = 100;
static const nat4 kMaxLeavesPerTree   = 1000;
static const nat4 kNumConnections     = 4;
static const nat4 kNumTrees           = 12;
static const nat4 kTooManyBugsPerLeaf = 50;

//--- Leaf Class --------------------------------------------------------------

class Leaf : public object {
 protected :
   nat2 mAlpha;
   nat2 mBlue;
   nat2 mGreen;
   nat4 mNumBugs;
   nat2 mRed;

 public :
   Leaf(nat2 iRed, nat2 iGreen, nat2 iBlue, nat2 iAlpha, nat4 iNumBugs);
   ~Leaf(void);

   bool needsAnExterminator(void) const;

   static void NeedsAnExterminator(ref<Leaf> iLeaf, const void *iArg);

   METACLASS_DECLARATIONS(Leaf, object);
};

REGISTER(Leaf, object, pessimistic_repeatable_read_scheme);

/*****************************************************************************\
Leaf - expected constructor
\*****************************************************************************/

Leaf::Leaf(nat2 iRed, nat2 iGreen, nat2 iBlue, nat2 iAlpha, nat4 iNumBugs) :
 object(self_class),
 mAlpha(iAlpha),
 mBlue(iBlue),
 mGreen(iGreen),
 mNumBugs(iNumBugs),
 mRed(iRed)
{
}

/*****************************************************************************\
~Leaf - destructor
\*****************************************************************************/

Leaf::~Leaf(void)
{
}


field_descriptor &Leaf::describe_components()
{
 return NO_FIELDS;
}


/*****************************************************************************\
needsAnExterminator - determine if this leaf has too many bugs on it
\*****************************************************************************/

bool Leaf::needsAnExterminator(void) const
{
 return (mNumBugs >= kTooManyBugsPerLeaf);
}


/*****************************************************************************\
NeedsAnExterminator - determine if that leaf has too many bugs on it
\*****************************************************************************/

void Leaf::NeedsAnExterminator(ref<Leaf> iLeaf, const void *iArg)
{
 if(iLeaf->needsAnExterminator()) {
   ref<Leaf> *leaf = (ref<Leaf> *)iArg;
   *leaf = iLeaf;
 }
}


//--- Tree Class --------------------------------------------------------------

class Tree : public B_tree {
 public :
   Tree(void);
   ~Tree(void);

   void leave(void);
   bool needsAnExterminator(void) const;
   void pickBuggyLeaf(void);
   void simulateSpring(void);

   METACLASS_DECLARATIONS(Tree, B_tree);
};

REGISTER(Tree, B_tree, pessimistic_repeatable_read_scheme);

/*****************************************************************************\
Tree - default constructor
\*****************************************************************************/

Tree::Tree(void) :
 B_tree(self_class)
{
}


/*****************************************************************************\
~Tree - destructor
\*****************************************************************************/

Tree::~Tree(void)
{
}


field_descriptor &Tree::describe_components()
{
 return NO_FIELDS;
}


/*****************************************************************************\
leave - drop all of the leaves from this tree
\*****************************************************************************/

void Tree::leave(void)
{
 ref<set_member> leaf = first;

 while(!leaf.is_nil()) {
   remove(leaf);
   leaf = first;
 }
}


bool Tree::needsAnExterminator(void) const
{
 ref<Leaf> buggyLeaf;
 iterate((member_function)Leaf::NeedsAnExterminator, &buggyLeaf);
 return !buggyLeaf.is_nil();
}


void Tree::pickBuggyLeaf(void)
{
 ref<Leaf> buggyLeaf;
 iterate((member_function)Leaf::NeedsAnExterminator, &buggyLeaf);
 if(!buggyLeaf.is_nil()) {
   remove(buggyLeaf);
 }
}


void Tree::simulateSpring(void)
{
 nat4 numLeaves = (nat4)rand() % kMaxLeavesPerTree;
 for(nat4 i = 0; i < numLeaves; i++) {
   char key[20];
   ref<Leaf> leaf = new Leaf(0,0xFFFF,0,0xFFFF, rand() % kMaxBugsPerLeaf);
   sprintf(key, "%08lx", i);
   insert(key, leaf);
 }
}


//--- Wood Class --------------------------------------------------------------

class Wood : public B_tree {
 public :
   Wood(void);
   ~Wood(void);

   ref<Tree> getTree(const char *iKey) const;
   void      initialize() const;

   METACLASS_DECLARATIONS(Wood, B_tree);
};

REGISTER(Wood, B_tree, pessimistic_repeatable_read_scheme);

/*****************************************************************************\
Wood - default constructor
\*****************************************************************************/

Wood::Wood(void) :
 B_tree(self_class)
{
}


/*****************************************************************************\
~Wood - destructor
\*****************************************************************************/

Wood::~Wood(void)
{
}


field_descriptor &Wood::describe_components()
{
 return NO_FIELDS;
}


ref<Tree> Wood::getTree(const char *iKey) const
{
 ref<Tree>       tree;
 ref<set_member> treeIndex;

 treeIndex = find(iKey);
 if(!treeIndex.is_nil())
   tree = treeIndex->obj;
 return tree;
}


void Wood::initialize() const
{
 if(is_abstract_root()) {
   ref<Wood> wood = this;
   modify(wood)->become(NEW Wood);
 }
}


database  db;
database *dbs[kNumConnections];
bool      quittingTime = false;

void ChooseTree(char *oKey)
{
 nat4 i = rand() % kNumTrees;
 sprintf(oKey, "%08lx", i);
}


void LeafPicker(void *arg)
{
 dbs[(nat4)arg]->attach();
 while(!quittingTime) {
   char            key[20];
   ref<Tree>       tree;
   ref<Wood>       wood;

   dbs[(nat4)arg]->get_root(wood);
   ChooseTree(key);
   tree = wood->getTree(key);
   if(tree->needsAnExterminator()) {
     modify(tree)->pickBuggyLeaf();
     fprintf(stderr, "LeafPicker %08lx picked a leaf off of \"%s\".\n",
             (nat4)task::current(), key);
   } else {
     fprintf(stderr, "LeafPicker %08lx skipped tree \"%s\" - it's okay.\n",
             (nat4)task::current(), key);
   }
 }
}


void Supervisor(void *)
{
 while(!quittingTime) {
   char            key[20];
   ref<Tree>       tree;
   ref<Wood>       wood;

   db.get_root(wood);
   ChooseTree(key);
   tree = wood->getTree(key);
   if(tree->needsAnExterminator()) {
     fprintf(stderr, "Supervisor %08lx found a buggy tree \"%s\".\n",
             (nat4)task::current(), key);
   } else {
     fprintf(stderr, "Supervisor %08lx thinks tree \"%s\" is okay.\n",
             (nat4)task::current(), key);
   }
   task::sleep(1);
 }
}


int main(int, char **)
{
     nat4      i;

 task::initialize(task::normal_stack);

 // Open the default database connection.
 if(!db.open("cnxnPool.cfg")) {
   fprintf(stderr, "Unable to connect to the database.\n\n");
   return 0;
 }

 // Create a wooded forest, in which we'll be playing.
 ref<Wood> wood;
 db.get_root(wood);
 wood->initialize();
 for(i = 0; i < kNumTrees; i++) {
   char key[20];
   sprintf(key, "%08lx", i);
   ref<Tree> tree = new Tree;
   modify(tree)->simulateSpring();
   modify(wood)->insert(key, tree);
 }

 // Initialize the database connection pool.
 for(i = 0; i < kNumConnections; i++) {
   dbs[i]  = new database;
   dbs[i]->open("cnxnPool.cfg");
 }

 // Start the threads that are randomly reading objects from the database.
 task::create((task::fptr)Supervisor);
 task::create((task::fptr)Supervisor);

 // Start the threads that are randomly modifying database objects using 
 // connections from the pool.
 for(i = 0; i < kNumConnections; i++) {
   task::create((task::fptr)LeafPicker, (void *)i);
 }

 // Let the test run for a while, then signal the workers to quit.
 task::sleep(300);
 quittingTime = true;

 return 0;
}
