//-< TESTCLU.CXX >---------------------------------------------------*--------*
// GOODS                     Version 1.0         (c) 1997  GARRET    *     ?  *
// (Generic Object Oriented Database System)                         *   /\|  *
//                                                                   *  /  \  *
//                          Created:     22-Jan-2001  K.A. Knizhnik  * / [] \ *
//                          Last update: 22-Jan-2001  K.A. Knizhnik  * GARRET *
//-------------------------------------------------------------------*--------*
// Test of object clustering: fetch of object clusters and bulk allocs
//-------------------------------------------------------------------*--------*

#include "goods.h"
#include "dbscls.h"

USE_GOODS_NAMESPACE

#define N_RECORDS    10000
#define N_COMPONENTS 10

class component : public object { 
  public:
    ref<component> next;
    int4           value;

    METACLASS_DECLARATIONS(component, object);

    component(int index, ref<component> chain) : object(self_class) { 
        next = chain;
        value = index;
    }
};

class record : public object { 
  public: 
    ref<set_member> key;
    ref<component>  components;

    METACLASS_DECLARATIONS(record, object);

    record(const char* key_str, int key_len) : object(self_class) { 
        key = set_member::create(this, key_str, key_len);
    }
};

class wood : public object { 
  public: 
    ref<B_tree> tree;
    
    void initialize() const { 
        if (is_abstract_root()) { 
            ref<wood> root = this;
            modify(root)->become(NEW wood());
        } 
    }

    void insert(nat8 key) const { 
        record* rec = NEW record((char*)&key, sizeof key);
        for (int i = N_COMPONENTS; --i >= 0;) { 
            rec->components = NEW component(i, rec->components);
        }
        modify(tree)->insert(rec->key);
    }

    void locate(nat8 key) const { 
        ref<set_member> mbr = tree->find((char*)&key, 
                                         sizeof key, 
                                         set_member::str2key((char*)&key, sizeof(key)));
        assert(!mbr.is_nil());
        ref<record> rec = mbr->obj;
        int sum = 0;
        for (ref<component> cmp = rec->components; !cmp.is_nil(); cmp = cmp->next) { 
            sum += cmp->value;
        }
        assert(sum == N_COMPONENTS*(N_COMPONENTS-1)/2);
    }
    
    METACLASS_DECLARATIONS(wood, object);

  protected: 
    wood() : object(self_class) { 
        tree = B_tree::create(this); 
    }
};
    

field_descriptor& record::describe_components()
{
    return FIELD(key), FIELD(components);
}

field_descriptor& component::describe_components()
{
    return FIELD(next), FIELD(value);
}

field_descriptor& wood::describe_components()
{
    return FIELD(tree);
}

REGISTER(wood, object, pessimistic_repeatable_read_scheme);
REGISTER(component, object, optimistic_scheme);
REGISTER(record, object, optimistic_scheme);

int main() 
{
    database db;
    task::initialize(task::normal_stack);

    db.enable_clustering(True);
    db.set_alloc_buffer_size(100);

    if (db.open("large.cfg")) { 
        int i, j, k, step;
        ref<wood> root;
        db.get_root(root);
        root->initialize();
        time_t start = time(NULL), last = start, curr;
        nat8 key = 1998;
        for (i = 0, j = 0, k = 0, step = 100; i < N_RECORDS; i++) { 
            key = (3141592621u*key + 2718281829u) % 1000000007u;
            root->insert(key);
            if (++j == step) {  
                curr = time(NULL);
                k += j;
                j = 0;
                if (curr != last) { 
                    printf("Height %d, %d records, %.3f records per second (%.3f for last %d)   \r", 
                           root->tree->get_height(), i+1, (double)(i+1) / (curr - start), 
                           (double)k / (curr - last), k);
                    k = 0;
                    last = curr;
                } else { 
                    step *= 10;
                }
            }
        }
        curr = time(NULL);
        printf("\nEllapsed time %d for %d records insertion (%.2f records per second)\n", 
               int(curr-start), N_RECORDS, (double)N_RECORDS / (curr-start)); 
        key = 1998;
        start = time(NULL);
        for (i = 0; i < N_RECORDS; i++) { 
            key = (3141592621u*key + 2718281829u) % 1000000007u;
            root->locate(key);
        }
        curr = time(NULL);
        printf("Ellapsed time %d for locating %d records (%.2f records per second)\n", 
               int(curr-start), N_RECORDS, (double)N_RECORDS / (curr-start)); 
        root->tree->ok();
        db.close();
        return EXIT_SUCCESS;
    } else { 
        console::output("Failed to open database\n");
        return EXIT_FAILURE;
    }
}


